-- channel package of helper functions to interact with the FPGA soft processor
--
-- Uses the ring_buffers and connctrl packages
--
-- API Functions:
--
--


local cc = require 'connctrl'
local ffi = require 'ffi'
local rb = require 'ring_buffers'
local config = require 'system_config'
local log = require 'log'

local channel = {}

local f = assert(io.open("/usr/include/fpga-firmware/common_structures.h", "r"))
local h = f:read("*all")
f:close()
ffi.cdef(h)

local function errorf(fmt, ...)
    error(string.format(fmt, ...), 2)
end


function channel:new_channel(devname, firmware_file_name, cmd_ct, irq_ct)
    local new_chan = {}
    local bar_map
    if config.device_type ~= 'edaqxr' then
        bar_map = cc.get_controller_map(devname)
        cc.controller_reset(devname)
        cc.load_microcode(devname,
            "/usr/share/fpga-firmware/"..firmware_file_name)
    else
        cc.controller_reset(devname)
    end
    cc.clear_interrupts(devname)

    if not irq_ct then irq_ct = "irq_s" end
    new_chan.devname = devname
    new_chan.command_rb = rb.make_new_bd_ring(devname, cmd_ct, cmd_ct, 256)
    new_chan.response_rb = rb.make_new_bd_ring(devname, irq_ct, irq_ct, 4096,
        new_chan.command_rb.total_size)

    new_chan.ring_buffer_table = {}
    --new_chan.ring_buffer_table['Command'] = self.command_rb
    --new_chan.ring_buffer_table['Interrupt'] = self.response_rb

    -- Tell the channel controller firmware where to set up DMA in phys mem
    if config.device_type ~= 'edaqxr' then
        bar_map:w32(0x14, new_chan.command_rb:get_initial_pa())
        bar_map:w32(0x18, new_chan.response_rb:get_initial_pa())
    else
        self.ring_buffers = {
            new_chan.command_rb,
            new_chan.response_rb,
        }
        cc.set_response_bd_ptr(devname, new_chan.response_rb.dma_map_pa)
    end

    self.cmd_id = 0

    cc.release_controller_reset(devname)

    setmetatable(new_chan, self)
    self.__index = self

    return new_chan
end

function channel:poll(proxy)
    local statusbuf, element_count, heartbeat, bytes_written =
        self.response_rb:get_buffer_ptr()

    if not statusbuf then return nil end

    local seqnum = bit.rshift(tonumber(statusbuf.status), 8)
    statusbuf.status = bit.band(tonumber(statusbuf.status), 0xff)
    if self.expected_seqnum and self.expected_seqnum ~= seqnum then
        if not config.warn_on_unexpected_fpga_sequence_number then
            errorf('%s SEQUENCE NUMBER ERROR, expected %d, got %d\n',
                self.devname, self.expected_seqnum, seqnum)
        else
            log:log(log.WARNING,
                string.format('%s SEQUENCE NUMBER ERROR, expected %d, got %d\n',
                    self.devname, self.expected_seqnum, seqnum))
            self.expected_seqnum = bit.band(self.expected_seqnum + 1, 0xffffff)
            return
        end
    end
    self.expected_seqnum = bit.band(seqnum + 1, 0xffffff)

    if statusbuf.status == ffi.C.CMD_UPDATE then
        local t = {} -- KLUDGE table allocation!!!
        t.cmd = tonumber(statusbuf.cmd)
        t.pa = tonumber(statusbuf.prm.phys_address)
        local bd_ring = assert(self.ring_buffer_table[t.pa])
        t.bufdata, t.element_count, t.heartbeat = bd_ring:get_buffer_ptr()
        assert(t.bufdata)
        proxy:buffer(t)
        bd_ring:free_buffer_ptr()
        self.response_rb:free_buffer_ptr()
        return true
    end

    local ret_tbl = {}
    ret_tbl.status = tonumber(statusbuf.status)
    ret_tbl.cmd = tonumber(statusbuf.cmd)
    ret_tbl.cmd_id = tonumber(statusbuf.cmd_id)

    if statusbuf.status == ffi.C.CMD_OK then
        -- OK
    elseif statusbuf.status == ffi.C.CMD_OK_TS then
        ret_tbl.ts = statusbuf.prm.ts
    elseif statusbuf.status == ffi.C.CMD_STARTED then
        ret_tbl.pa = tonumber(statusbuf.prm.phys_address)
    elseif statusbuf.status == ffi.C.CMD_STOPPED then
        ret_tbl.pa = tonumber(statusbuf.prm.phys_address)
        self:delete_ring_buffer(ret_tbl.pa)
    elseif statusbuf.status == ffi.C.CMD_CHANNEL_SPECIFIC then
        ret_tbl.channel_specific = statusbuf.prm.channel_specific
    else
        local errmsg = ffi.string(statusbuf.prm.error_message,
            bytes_written - 8)
        local err_type = (statusbuf.status == ffi.C.CMD_ERROR)
            and 'COMMAND' or 'ASYNC'
        self.response_rb:free_buffer_ptr()
        error(self.devname .. ': ' .. err_type .. ' ERROR - ' .. errmsg)
    end

    self.response_rb:free_buffer_ptr()
    return ret_tbl
end

function channel:send_command_buffer(buf, size_override)
    self.cmd_id = self.cmd_id + 1
    buf.cmd_id = self.cmd_id
    self.command_rb:put_buffer(buf, size_override)
    return self.cmd_id
end

function channel:send_buffer(buf, name, size_override)
    assert(self.ring_buffer_table[name]:put_buffer(buf, size_override))
end

function channel:poll_response_buffer()
    local buf, bytes_written = self.response_rb:get_buffer()
    if not buf then return end
    local seqnum = bit.rshift(tonumber(buf.status), 8)
    buf.status = bit.band(tonumber(buf.status), 0xff)
    if self.expected_seqnum and self.expected_seqnum ~= seqnum then
        errorf('%s SEQUENCE NUMBER ERROR, expected %d, got %d\n',
            self.devname, self.expected_seqnum, seqnum)
    end
    self.expected_seqnum = bit.band(seqnum + 1, 0xffffff)
    return buf, bytes_written
end

function channel:poll_buffer(name)
    local statusbuf, bytes_written, element_count, heartbeat =
        assert(self.ring_buffer_table[name]):get_buffer()
    return statusbuf, bytes_written, element_count, heartbeat
end

function channel:remove_aux_ring_buffers()
    if self.ring_buffers then
        while #self.ring_buffers > 2 do
            table.remove(self.ring_buffers)
        end
    end
end

function channel:add_ring_buffer(name, buffer_ct, element_ct, count)
    local offset
    if self.ring_buffers then
        for _, v in ipairs(self.ring_buffers) do
            offset = offset or 0
            offset = offset + v.total_size
        end
    end
    local rbuf =
        rb.make_new_bd_ring(self.devname, buffer_ct, element_ct, count, offset)
    if self.ring_buffers then
        self.ring_buffers[#self.ring_buffers + 1] = rbuf
    end
    rbuf.name = name
    self.ring_buffer_table[name] = rbuf
    self.ring_buffer_table[rbuf:get_initial_pa()] = rbuf
    return rbuf:get_initial_pa()
end

function channel:get_ring_buffer(name)
    return self.ring_buffer_table[name]
end

function channel:delete_ring_buffer(name)
    local rbuf = self.ring_buffer_table[name]
    if rbuf == nil then return end
    self.ring_buffer_table[rbuf.name] = nil
    self.ring_buffer_table[rbuf:get_initial_pa()] = nil
    rb.destroy(rbuf)
end

function channel:calculate_scg_parameters(rate_prms, rate_num, rate_dem, clk)
    rate_prms.base_period = math.floor((clk * rate_dem) / rate_num)
    rate_prms.rate_numerator = rate_num
    rate_prms.remainder = (clk * rate_dem) % rate_num
end

function channel:int_ack()
    cc.controller_reset(self.devname)
end

function channel:wait_for_interrupt()
    cc.wait_for_interrupt(self.devname)
end

function channel:halt(dump_trace)
    cc.controller_reset(self.devname)
    if dump_trace then cc.dump_trace_buffer(self.devname) end
end

function channel:dump_ring_buffers()
    for k,v in pairs(self.ring_buffer_table) do
        if type(v) == "table" then
            v:dump(k)
        end
    end
end

return channel
