require 'strict'
local config = require 'system_config'
local log = require 'log'

log:open(arg[0], log.USER)

local args = {}

for _, a in ipairs(arg) do args[#args + 1] = a end

local fw_updater_type = arg[0]:match('.*/(.*)-')

local dbus_module = require 'dbus'
local ffi = require 'ffi'

local htonl = ffi.abi("le") and bit.bswap or function(x) return x end
local ntohl = htonl

local C = ffi.C

ffi.cdef[[
    typedef int ssize_t;
    typedef unsigned int        mode_t;
    typedef int pid_t;

    int setenv(const char *name, const char *value, int overwrite);
    int unsetenv(const char *name);
    unsigned int sleep(unsigned int seconds);
    int open(const char *pathname, int flags, mode_t mode);
    int close(int fd);
    ssize_t read(int fd, void *buf, size_t count);
    ssize_t write(int fd, const void *buf, size_t count);
    int creat(const char *pathname, mode_t mode);
    int unlink(const char *pathname);
    pid_t getpid();

    enum {
        O_RDONLY = 0,
        O_WRONLY = 1,
        O_RDWR = 2,
        O_CREAT = 0100,
    };
]]

local function cmd(fmt, ...)
    local command = string.format(fmt, ...)
    io.stderr:write(string.format('%s\n', command))
    local got = os.execute(command) -- KLUDGE
    io.stderr:write(string.format('...done: %s\n', tostring(got)))
end

local function popen_cmd(fmt, ...)
    local command = string.format(fmt, ...)
    local f = io.popen(command)
    if f then
        for line in f:lines() do
            io.stdout:write(string.format('%s\n', line))
            io.stdout:flush()
        end
        f:close()
    end
end

local dbus = dbus_module.bus_get('system')
dbus:hook_libev_default_loop()
local network = assert(dbus:auto_proxy_wait('com.somat.armada.Network',
                                            '/com/somat/armada/Network'))

local units = {}

local passwd = {
    edaq = 'root',
    quantum = 'hbmquantumx'
}

local logname = {
    edaq = 'root',
    quantum = 'hbmuser'
}

if fw_updater_type == 'edaq' then
    local rr = network:QueryRR('*', 'A', '*', '*')
    C.setenv('DROPBEAR_PASSWORD', passwd[fw_updater_type], 1)
    for _, v in ipairs(rr) do
        local arch = ''
        local pcmd = string.format("dbclient -y 'root@%s' uname -m", v[3])
        local fh = io.popen(pcmd)
        if fh then
            arch = fh:read('*a'):gsub('%s*$', '')
            fh:close()
        end
        local unit = {name=v[1], ip=v[3], arch=arch}
        units[unit.name] = unit
        print(unit.name, unit.arch, unit.ip)
    end
elseif fw_updater_type == 'quantum' then
    local rr = network:QueryRR('*', 'A', '*', '*')
    for _, v in ipairs(rr) do
        for i = 2, #args do
            if v[1] == args[i] then
                units[args[i]] = {
                    name = v[1],
                    ip = v[3],
                    arch = 'ppc'
                }
            end
        end
    end
    -- Check for devices specified by IPv6 address as well
    for i = 2, #args do
        if args[i]:match('^%x+::%x+:%x+:') then
            units[args[i]] = {
                name = args[i],
                ip = args[i],
                arch = 'ppc'
            }
            -- Do any fix ups for link local address
            if args[i]:match('^fe80::') and not args[i]:match('%ethint') then
                units[args[i]].ip = units[args[i]].ip .. '%ethint'
            end
        end
    end
elseif fw_updater_type == 'dx' then
    local rr = network:QueryRR('dx*', 'A', '*', '*')
    for _, v in ipairs(rr) do
        for i = 2, #args do
            if v[1] == args[i] then
                units[args[i]] = {
                    name = v[1],
                    ip = v[3],
                    arch = 'ppc'
                }
            end
        end
    end
else
    io.stderr:write(string.format(
        'Usage: %s-updater <fw-image> <device-name> ...\n', fw_updater_type))
    io.stderr:write(string.format('Invalid updater type\n'))
    io.stderr:write(string.format('Must be "edaq" or "quantum"\n'))
    os.exit(1)
end

if #args < 2 then
    io.stderr:write(string.format(
        'Usage: %s-updater <fw-image> <device-name> ...\n', fw_updater_type))
    os.exit(1)
end

local image = table.remove(args, 1)
if not image or #args == 0 then
    error(string.format('Usage: %s-updater <fw-image> <device-name> ...',
        fw_updater_type))
end


-- Function used to extract the firmware image from a HBG file
-- FIXME:  The current version of HBG files does NOT conform to the
-- specification and this function will have to change when it does.
local function extract_firmware(hbg_file)
    local fd = C.open(hbg_file, C.O_RDONLY, 0)
    local size = ffi.new('int[1]')
    local buf = ffi.new('unsigned char[256]')
    local states = {
        'get_handbag_key',
        'get_handbag_value',
        'get_file_name',
        'get_file',
    }
    local state = states[1]
    local image_name =
        string.format('/usr/share/quantum-firmware/image-%d.qxi', C.getpid())
    while true do
        local amount_read = C.read(fd, size, ffi.sizeof(size))
        if amount_read == ffi.sizeof(size) then
            size[0] = ntohl(size[0])
            if size[0] <= ffi.sizeof(buf) then
                amount_read = C.read(fd, buf, size[0])
                if amount_read ~= size[0] then
                    log:log(log.ERR, "Unable to read from handbag file")
                    C.close(fd)
                    return
                else
                    if state == states[1] then
                        if ffi.string(buf, size[0]) == 'handbag' then
                            state = states[2]
                        else
                            log:log(log.ERR, "Unexpected data in HBG file")
                            C.close(fd)
                            return
                        end
                    elseif state == states[2] then
                        -- I think this is the handbag version
                        local v = ffi.cast('int *', buf)
                        if ntohl(v[0]) ~= 1 then
                            log:log(log.ERR, "Unexpected HBG file version")
                            C.close(fd)
                            return
                        else
                            state = states[3]
                        end
                    elseif state == states[3] then
                        local fname = ffi.string(buf, size[0])
                        state = states[4]
                    end
                end
            elseif state == states[4] then
                local amount2read = size[0]
                -- We are positioned at the beginning of the file
                local fd2 = C.creat(image_name, 420)
                if fd2 < 0 then
                    log:log(log.ERR, "Unable to create temporary qxi file")
                    C.close(fd)
                    return
                end
                while amount2read > 0 do
                    local left = amount2read > 256 and 256 or amount2read
                    local amount_read = C.read(fd, buf, left)
                    if amount_read ~= left then
                        log:log(log.ERR, "Unable to read qxi from hbg")
                        C.close(fd)
                        C.close(fd2)
                        C.unlink(image_name)
                        return
                    else
                        local written = C.write(fd2, buf, amount_read)
                        if written ~= amount_read then
                            log:log(log.ERR, "Unable to write qxi")
                            C.close(fd)
                            C.close(fd2)
                            C.unlink(image_name)
                            return
                        end
                        amount2read = amount2read - amount_read
                    end
                end
                C.close(fd)
                C.close(fd2)
                return image_name
            end
        else
            log:log(log.ERR, "Unable to read size from handbag file")
            C.close(fd)
            return
        end
    end
end

for _, device_name in ipairs(args) do
    local unit = units[device_name]
    if fw_updater_type == 'edaq' and unit and unit.arch == 'i586' then
        cmd("rsync -P -e 'dbclient -y' %s 'root@[%s]:/'", image, unit.ip)
        cmd("dbclient -y 'root@%s' /sbin/reboot", unit.ip)
        C.sleep(5)
        -- KLUDGE do some stuff here to figure out when the eDAQ comes back up
--        cmd("ping6 %s", unit.ip)
    elseif fw_updater_type == 'quantum' and unit and unit.arch == 'ppc' then
        log:log(log.INFO, string.format("Update firmware on %s", args[1]))

        -- KLUDGE: Kill off the units HBM discovery announcements while the
        -- firmware update is occurring!  This prevents the MX manager from
        -- detecting a device online without all of it's services and rebooting
        -- it.
        local cmd = string.format(
            "DROPBEAR_PASSWORD=hbmquantumx /usr/bin/dbclient -y hbmuser@%s " ..
            "'killall devscan.bin'", unit.ip)
        os.execute(cmd)

        if image:find("%.hbg") then
            image = extract_firmware(image)
            if not image then
                log:log(log.ERR, "Unable to extract qxi file from hbg file")
                return
            end
            if config.device_type == 'cx23r' then
                popen_cmd("curl -N -s -X POST -T %s " ..
                    "http://%s/update/firmware?force=true", image, unit.ip)
            else
                popen_cmd("curl -N -s --interface ethint -X POST -T %s " ..
                    "http://[%s]/update/firmware?force=true", image,
                    unit.ip:gsub('%%ethint', ''))
            end
            C.unlink(image)
        else
            popen_cmd("curl -N -s -X POST -T %s " ..
                "http://%s/update/firmware?force=true", image, unit.ip)
        end
        if config.device_type == 'cx23r' then
            os.execute(string.format("curl -N -s -X POST http://%s/reboot",
                unit.ip))
        else
            os.execute(string.format(
                "curl -N -s --interface ethint -X POST http://[%s]/reboot",
                unit.ip:gsub('%%ethint', '')))
        end
        log:log(log.INFO,
            string.format("Firmware update on %s complete", args[1]))
    elseif fw_updater_type == 'dx' then
        log:log(log.INFO, string.format("Update firmware on %s", unit.name))
        if unit.name:find('dx23r') then
            os.execute(string.format('DROPBEAR_PASSWORD=root dbclient -y root@%s "/usr/bin/led-control eight_hz green"', unit.ip))
            os.execute(string.format('DROPBEAR_PASSWORD=root scp -S /usr/bin/scp-script /usr/share/dx23r-firmware/__ng-update.tar.gz root@[%s]:/', unit.ip))
        end
        os.execute(string.format('DROPBEAR_PASSWORD=root dbclient -y root@%s "sync;sync;reboot"', unit.ip))
        log:log(log.INFO,
            string.format("Firmware update on %s complete", unit.name))
    end
end
