require('lj_include'){
    'stdlib',
    'string',
}

local bit = require 'bit'
local ffi = require 'ffi'

local C = ffi.C

local bit_bor = bit.bor
local bit_rshift = bit.rshift
local bit_tobit = bit.tobit
local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_fill = ffi.fill
local ffi_istype = ffi.istype
local ffi_string = ffi.string
local math_max = math.max

ffi.cdef[[
    struct membuf {
        size_t size;
        size_t max_size;
        uint8_t *buf;
    };
]]

local function errorf(fmt, ...)
    error(string.format(fmt, ...))
end

local mt = {}

local mt_ctype

local function p_size(p, size)
    if ffi_istype(mt_ctype, p) then
        return p.buf, (size or p.size)
    elseif p and size then
        return p, size
    elseif not size then
        p = tostring(p)
        return p, #p
    else
        errorf("Bad arguments p=%s, size=%s", p, size)
    end
end

function mt:initialize(max_size, init)
    self:expand(max_size)
    if init then
        self:append(init)
    end
    return self
end

function mt:head()
    return self.buf
end

function mt:tail()
    return self.buf + self.size
end

function mt:available()
    return self.max_size - self.size
end

function mt:assure_available()
    if self:available() <= 0 then
        self:expand(self.max_size * 2)
    end
    return self
end

local function round_up_power_of_2(v)
    -- See: http://graphics.stanford.edu/~seander/bithacks.html#RoundUpPowerOf2
    v = bit_tobit(math_max(v, 1) - 1)
    v = bit_bor(v, bit_rshift(v, 1))
    v = bit_bor(v, bit_rshift(v, 2))
    v = bit_bor(v, bit_rshift(v, 4))
    v = bit_bor(v, bit_rshift(v, 8))
    v = bit_bor(v, bit_rshift(v, 16))
    return v + 1
end

function mt:expand(max_size)
    if max_size <= self.max_size then
        return self
    end
    local size = round_up_power_of_2(max_size)
    local buf = C.realloc(self.buf, size)
    if buf == nil then
        errorf("realloc() of buf=%s size=%s to size=%s (max_size=%s) failed\n",
            self.buf, self.max_size, size, max_size)
    end
    self.buf = buf
    self.max_size = size
    ffi_fill(self:tail(), self:available())
    return self
end

function mt:append(p, size)
    if p ~= nil then
        p, size = p_size(p, size)
    end
    self:expand(self.size + size)
    if p ~= nil then
        ffi_copy(self:tail(), p, size)
    end
    self.size = self.size + size
    return self
end

function mt:remove(pos, len)
    local tomove = self.size - pos - len
    if tomove < 0 then
        errorf("Bad arguments to membuf:remove() pos=%s + len=%s > size=%s",
            pos, len, self.size)
    end
    if tomove > 0 then
        C.memmove(self.buf + pos, self.buf + pos + len, tomove)
    end
    self.size = self.size - len
    return self
end

function mt:extract(pos, len)
    local str = ffi_string(self.buf + pos, len)
    self:remove(pos, len)
    return str
end

function mt:clear()
    self.size = 0
    return self
end

function mt:search(offset, needle, len)
    needle, len = p_size(needle, len)
    if offset + len > self.size then
        return nil
    end
    local p = ffi_cast('uint8_t *',
        C.memmem(self.buf + offset, self.size - offset, needle, len))
    if p == nil then
        return nil
    end
    return p - self.buf
end

function mt:__len()
    return self.size
end

function mt:__tostring()
    return ffi_string(self.buf, self.size)
end

function mt:__concat(a)
    return self:append(a)
end

function mt:__gc()
    C.free(self.buf)
    self.buf = nil
end

mt.__index = mt

mt_ctype = ffi.metatype('struct membuf', mt)

local m = {}

function m:new(init, max_size)
    if not max_size then
        if type(init) == 'number' then
            max_size = init
            init = nil
        else
            max_size = #init
        end
    end
    local buf = mt_ctype()
    return buf:initialize(max_size, init)
end

return m
