--
-- Copyright (C) 2016 HBM, Inc. All rights reserved.
--

local calibration   = require 'calibration'
local channel_proxy = require 'channel_proxy'
local mx_can        = require 'mx.can'
local task_chain    = require 'task_chain'

local mxChannelProxy = class('mxChannelProxy', channel_proxy)

function mxChannelProxy:initialize(mxDevice, channel_num, info)

    self.mxdev              = mxDevice
    self.proxy_type         = 'mxChannelProxy'
    self.can_info           = info
    self.num_signals        = 0
    self.signal_slots       = {}
    self.queued_channels    = {}

end

function mxChannelProxy:available_channels(hostname)

    return {
        {
            name        = 'raw_' .. hostname:gsub('%W', '_'),
            connector   = '@' .. hostname,
            description = 'CAN messages',
            data_type   = 'message_float64',
            category    = 'source/hbm_mx471b',
            sample_rate = 10
        }
    }

end

function mxChannelProxy:start_comet_stream(w, chans, sink)

end

function mxChannelProxy:start_channel_stream(w, spec_array, md_sink, sink)

    for _, spec in ipairs(spec_array) do

        -- If this is a new test ID, discard any queued channels.
        self.mxdev.queued_channels = self.mxdev.queued_channels or {}
        if spec._test_id ~= self.mxdev.test_id then
            local virt_chans = self.mxdev.mx.control.device.virtual_channels
            for _, virt_chan in ipairs(virt_chans) do
                virt_chan.proxy.queued_channels = {}
            end
            self.mxdev.queued_channels = {}
            self.mxdev.test_id = spec._test_id
        end

        -- Queue up all channel requests until all of them have been received.
        -- We're maintaining TWO lists here: self.queued_channels, which has all
        -- channels for a given CAN connector, and self.mxdev.queued_channels,
        -- which is a global queue of all channels in the test for this MX471B
        -- device. We need the latter to enforce the 128-per-connector limit, and
        -- we need the former to know when we've received start_channel_stream()
        -- calls for all the channels so we can start up.

        local queued_channel = {
            w                   = w,
            spec                = spec,
            sink                = sink,
            md_sink             = md_sink,
            stream_reference    = string.format(
                                    'CanReceiver_Connector%d_Decoder%d.Signal1',
                                    self.can_info.channel_num,
                                    #self.queued_channels + 1)
        }

        self.queued_channels[#self.queued_channels + 1] = queued_channel
        self.mxdev.queued_channels[#self.mxdev.queued_channels + 1] =
            queued_channel

        if #self.mxdev.queued_channels == spec._channel_count then
            self:_prepare_test(md_sink)
        end

    end

    -- When the test ends, detach all channel sinks and tell libmxstream
    -- to unsubscribe from all of the signals.
    w.cleanup = function()
        self.mxdev.active_channels = self.mxdev.active_channels or 0
        self.mxdev.active_channels = self.mxdev.active_channels - #spec_array
        if self.mxdev.active_channels == 0 then
            local tasks = task_chain:new()
            for i, channel in ipairs(self.mxdev.queued_channels) do
                self.mxdev.mx.hbmstream:remove_subscription(
                    channel.stream_reference)
            end
            self.mxdev.mx.hbmstream:update_subscriptions()
            local decoder_nums = {}
            for i = 1, 128 do decoder_nums[i] = i end
            for i = 1, 4 do
                tasks:add(function(on_success, on_error)
                    return mx_can.activate_all_decoders(self.mxdev,
                        i, decoder_nums, '0', on_success, on_error) end)
            end
            tasks:run()
        end
    end

end

function mxChannelProxy:_prepare_test(md_sink)

    -- Validate setup for each of the four CAN connectors
    local virt_chans = self.mxdev.mx.control.device.virtual_channels
    for _, virt_chan in ipairs(virt_chans) do

        -- Only support up to 128 signal slots per CAN connector
        local chans = virt_chan.proxy.queued_channels or {}
        if #chans > 128 then
            md_sink({ error_message = "Cannot support more than " ..
                "128 CAN signals on a single MX471B channel" })
            return
        end

        if #chans > 0 then
            -- All signals on a connector must have same bitrate/termination
            for _, channel in ipairs(chans) do
                if channel.spec.bit_rate ~= chans[1].spec.bit_rate then
                    md_sink({ error_message = "CAN signals on a single " ..
                        "connector must have the same bit rate" })
                    return
                end
                if channel.spec.internal_termination ~=
                    chans[1].spec.internal_termination then
                    md_sink({ error_message = "CAN signals on a single " ..
                        "connector must have the same termination setting" })
                    return
                end
            end
        end

    end

    -- Set up a chain of tasks/callbacks to perform.
    local tasks = task_chain:new()
    local xml = nil

    local virt_chans = self.mxdev.mx.control.device.virtual_channels

    -- Send XML configuration for the baud rate and termination settings
    -- for each enabled connector, one by one.
    for connector_num, virt_chan in ipairs(virt_chans) do
        for _, chan in ipairs(virt_chan.proxy.queued_channels) do
            tasks:add(function(on_success, on_error)
                return mx_can.set_canbus_settings(self.mxdev, connector_num,
                    chan.spec.bit_rate, chan.spec.internal_termination, nil,
                    on_success, on_error) end)
            break
        end
    end

    -- Send XML configuration for each channel.
    local activations = {}
    for connector_num, virt_chan in ipairs(virt_chans) do
        for decoder_num, chan in ipairs(virt_chan.proxy.queued_channels) do
            tasks:add(function(on_success, on_error)
                return mx_can.set_decoder_settings(self.mxdev, connector_num,
                    decoder_num, chan.spec, on_success, on_error) end)
            tasks:add(function(on_success, on_error)
                return mx_can.set_async_signal_settings(self.mxdev,
                    connector_num, decoder_num, on_success, on_error) end)
            activations[connector_num] = activations[connector_num] or {}
            activations[connector_num][decoder_num] = decoder_num
        end
    end

    -- Set all used decoders to active.
    for connector_num, decoder_nums in pairs(activations) do
        tasks:add(function(on_success, on_error)
            return mx_can.activate_all_decoders(self.mxdev, connector_num,
                decoder_nums, '1', on_success, on_error) end)
    end

    -- The last task is to announce test readiness to the test engine
    -- for all connectors. When all channels have announced ready, the
    -- test engine will call the callback below to actually subscribe
    -- to the MX signals and begin data acquisition.
    tasks:add(
        function(on_success, on_error)
            for connector_num, virt_chan in ipairs(virt_chans) do
                for _, chan in ipairs(virt_chan.proxy.queued_channels) do
                    virt_chan.proxy:announce_ready(
                        chan.spec.connector, true, 'big')
                end
            end
        end)

    -- When all channels have announced ready, the test engine will call
    -- this function, at which time we will subscribe to the MX signals and
    -- begin data acquisition.
    self.mxdev.start_callback = function(_self, start_time)
        self:_start_test(md_sink)
    end

    -- Call the device start callback when the start time is announced
    self.start_callback = function(self, start_time)
        self.start_callback = nil
        self.mxdev.start_callback(self.mxdev, start_time)
    end

    tasks:run()

end

function mxChannelProxy:_start_test(md_sink)

    -- Set up a chain of tasks/callbacks to perform.
    local tasks = task_chain:new()

    -- Tell libhbmstream to subscribe to all of the channels.
    tasks:add(
        function(on_success, on_error)
            for _, channel in ipairs(self.mxdev.queued_channels) do
                self.mxdev.mx.hbmstream:add_subscription(
                    channel.stream_reference, channel.w.socket)
            end
            self.mxdev.mx.hbmstream:update_subscriptions()
        end)

    -- Initialize this for mx_proxy's benefit (though we won't be using
    -- the start_channel_list at all for our purposes).
    self.mxdev.start_channel_list = self.mxdev.start_channel_list or {}

    -- Run all the tasks we just queued up.
    self.mxdev.active_channels = #self.mxdev.queued_channels
    tasks:run()

end

return mxChannelProxy
