--------------------------------------------------------------------------------
-- Copyright (C) 2016 HBM, Inc. All rights reserved.
--------------------------------------------------------------------------------

require 'middleclass'

require('lj_include') {
    'errno',
    'fcntl',
    'string',
    'unistd',
}

local bit   = require 'bit'
local ev    = require 'ev'
local fifo  = require 'net/fifo'
local ffi   = require 'ffi'
local unix  = require 'unix'

local C = ffi.C

--------------------------------------------------------------------------------

local tcp_stream = class('tcp_stream')

--------------------------------------------------------------------------------

function tcp_stream:initialize(loop, socket)

    local ok, ex = pcall(function()

        self.loop = loop
        self.socket = socket
        self.rxq = fifo:new(16 * 1024, 1024 * 1024)
        self.txq = fifo:new(16 * 1024, 32 * 1024 * 1024)

        unix.setfl(self.socket, C.O_NONBLOCK)

        self.rx_watcher = ev.IO.new(function(...) self:_on_rx(...) end, self.socket, ev.READ)
        self.tx_watcher = ev.IO.new(function(...) self:_on_tx(...) end, self.socket, ev.WRITE)

        self.rx_watcher:start(self.loop)
        self.running = true

    end)

    if not ok then

        self:_shutdown(self)
        error(ex)

    end

end

--------------------------------------------------------------------------------

function tcp_stream:_on_rx(loop, watcher, e)

    local total = 0

    while true do

        local buf, size = self.rxq:alloc_add()
        if size == 0 then
            self:_error('RX overflow')
            return
        end

        local rv = C.recv(self.socket, buf, size, 0)

        if rv < 0 then
            local errno = ffi.errno()
            if errno == C.EAGAIN then break end
            self:_error('RX error', errno)
            return
        elseif rv == 0 then
            self:_shutdown()
            return
        else
            self.rxq:commit_add(rv)
            total = total + rv
        end

    end

    if total > 0 and self.on_data then self.on_data(self) end

end

--------------------------------------------------------------------------------

function tcp_stream:_on_tx(loop, watcher, e)

    if not self.running then return end

    local sent = 0
    local buf = self.txq:data()
    local size = self.txq:size()

    while sent < size do
        local rv = C.send(self.socket, buf + sent, size - sent, 0)
        if rv < 0 then
            local errno = ffi.errno()
            if errno == C.EAGAIN then break end
            self:_error('TX error', errno)
            return
        elseif rv == 0 then
            self:_shutdown()
            return
        else
            sent = sent + rv
        end
    end

    self.txq:consume(sent)

    if sent == size then
        watcher:stop(loop)
    end

end

--------------------------------------------------------------------------------

function tcp_stream:fifo()
    return self.rxq
end

--------------------------------------------------------------------------------

function tcp_stream:close()
    self:_shutdown()
end

--------------------------------------------------------------------------------

function tcp_stream:send(ptr, len)

    if not self.running then return end

    if type(ptr) == 'string' and not len then
        return self:send(ptr, ptr:len())
    end

    local p = ffi.cast('unsigned char *', ptr)
    local sent = 0

    if self.txq:size() == 0 then
        while sent < len do
            local rv = C.send(self.socket, p + sent, len - sent, 0)
            if rv < 0 then
                local errno = ffi.errno()
                if errno ~= C.EAGAIN then
                    self:_error('TX error', errno)
                    return
                end
                break
            elseif rv == 0 then
                self:_shutdown()
                break
            else
                sent = sent + rv
            end
        end
    end

    if sent < len then
        local rv = self.txq:add(p + sent, len - sent)
        if rv ~= len - sent then
            self:_error('TX overflow')
            return
        end
        self.tx_watcher:start(self.loop)
    end

end

--------------------------------------------------------------------------------

function tcp_stream:_shutdown()

    if self.socket then C.close(self.socket) end
    if self.rxq then self.rxq:destroy() end
    if self.txq then self.txq:destroy() end
    if self.rx_watcher then
        self.rx_watcher:stop(self.loop)
        self.rx_watcher:clear_pending(self.loop)
    end
    if self.tx_watcher then
        self.tx_watcher:stop(self.loop)
        self.tx_watcher:clear_pending(self.loop)
    end

    self.socket = nil
    self.rxq = nil
    self.txq = nil
    self.rx_watcher = nil
    self.tx_watcher = nil

    if self.running then
        self.running = false
        if self.on_close then self.on_close(self) end
    end

end

--------------------------------------------------------------------------------

function tcp_stream:_error(message, errno)
    if errno then message = message .. string.format(' (errno:% d)', errno) end
    if self.running then
        if self.on_error then self.on_error(self, message) end
    end
    self:_shutdown()
    return message
end

--------------------------------------------------------------------------------

function tcp_stream:on_data(handler) self.on_data = handler end
function tcp_stream:on_close(handler) self.on_close = handler end
function tcp_stream:on_error(handler) self.on_error = handler end

--------------------------------------------------------------------------------

return tcp_stream
