require('lj_include'){
    'string',
    'unistd',
    'sys/types',
    'sys/socket',
    'sys/ioctl',
    'net/if',
    'linux/sockios',
    'linux/mii',
}

local ffi = require 'ffi'
local C = ffi.C
local ev = require 'ev'
local bit = require 'bit'

local net = {}

net.ev = require 'ev'

local function file_exists(name)
    local f = io.open(name, 'r')
    if f ~= nil then
        io.close(f)
        return true
    end
end

function net.file_exists(name)
    return file_exists(name)
end

function net.get_info(ifname, fd)
    local info = {}
    local skfd = fd or C.socket(C.AF_INET, C.SOCK_DGRAM, 0)
    if skfd < 0 then
        error("Unable to create socket")
    end

    local ifr = ffi.new('struct ifreq[1]')

    local function mdio_read(location)
        local mii = ffi.cast('struct mii_ioctl_data *',
            ifr[0].ifr_ifru.ifru_slave)
        mii.reg_num = location
        if C.ioctl(skfd, C.SIOCGMIIREG, ifr) < 0 then
            local e = ffi.errno()
            C.close(skfd)
            error(string.format("SIOCGMIIREG on %s failed: %s",
                ifr[0].ifr_ifrn.ifrn_name, ffi.string(C.strerror(e))))
        end
        return mii.val_out
    end

    C.strncpy(ifr[0].ifr_ifrn.ifrn_name, ifname, C.IFNAMSIZ)
    if C.ioctl(skfd, C.SIOCGMIIPHY, ifr) < 0 then
        local e = ffi.errno()
        C.close(skfd)
        error(string.format("SIOCGMIIPHY on %s failed: %s",  ifname,
            ffi.string(C.strerror(e))))
        return
    end
    local status_reg = mdio_read(C.MII_BMSR)
    info.link_up = 'down'
    info.speed = 0
    info.duplex = 'n/a'
    if bit.band(status_reg, C.BMSR_LSTATUS) ~= 0 then
        info.link_up = 'up'
        info.speed = 100
        info.duplex = 'half'
        local control_reg = mdio_read(C.MII_BMCR)
        local sr_1000 = mdio_read(C.MII_STAT1000)
        local lpa_r = mdio_read(C.MII_LPA)
        if bit.band(sr_1000, C.MII_LPA_1000FULL + C.MII_LPA_1000HALF) ~= 0 then
            info.speed = 1000
            if bit.band(sr_1000, C.MII_LPA_1000FULL) then
                info.duplex = 'full'
            end
        elseif bit.band(lpa_r, C.MII_LPA_100FULL + C.MII_LPA_100HALF) ~= 0 then
            info.speed = 100
            if bit.band(lpa_r, C.MII_LPA_100FULL) then
                info.duplex = 'full'
            end
        elseif bit.band(lpa_r, C.MII_LPA_10FULL + C.MII_LPA_10HALF) ~= 0 then
            info.speed = 10
            if bit.band(lpa_r, C.MII_LPA_10FULL) then
                info.duplex = 'full'
            end
        elseif bit.band(control_reg, C.MII_BMCR_SPEED1000) ~= 0 then
            info.speed = 1000
            if bit.band(control_reg, C.MII_BMCR_FULLDPLX) ~= 0 then
                info.duplex = 'full'
            end
        end
    end

    if not fd then
        C.close(skfd)
    end
    return info
end

function net.carrier_up(iface, loop, callback)
    local skfd = C.socket(C.AF_INET, C.SOCK_DGRAM, 0)
    if skfd < 0 then
        error("Unable to create socket")
    end

    local ifr = ffi.new('struct ifreq[1]')

    local function mdio_read(location)
        local mii = ffi.cast('struct mii_ioctl_data *',
            ifr[0].ifr_ifru.ifru_slave)
        mii.reg_num = location
        if C.ioctl(skfd, C.SIOCGMIIREG, ifr) < 0 then
            local e = ffi.errno()
            C.close(skfd)
            error(string.format("SIOCGMIIREG on %s failed: %s",
                ifr[0].ifr_ifrn.ifrn_name, ffi.string(C.strerror(e))))
        end
        return mii.val_out
    end

    -- Function used to determine if an interface exists.  Note, we do up to 10
    -- retries at a rate of 1/second on this because the ioctl can fail with
    -- EINVAL if this function is used early in the system boot sequence.
    local function iface_exists(ifname)
        for retry = 1, 10 do
            C.strncpy(ifr[0].ifr_ifrn.ifrn_name, ifname, C.IFNAMSIZ)
            if C.ioctl(skfd, C.SIOCGMIIPHY, ifr) < 0 then
                local e = ffi.errno()
                if e == C.ENODEV then
                    return false
                elseif e == C.EINVAL then
                    C.sleep(1)
                else
                    C.close(skfd)
                    error(string.format("SIOCGMIIPHY on %s failed: %s",  ifname,
                        ffi.string(C.strerror(e))))
                    return
                end
            end
            return true
        end
        error(string.format(
            "Unable to determine if interace %s exists", ifname))
    end

    if not iface_exists(iface) then
        print(string.format("No such network interface: %s", iface))
        C.close(skfd)
        return
    end

    local function link_up(ifname)
        C.strncpy(ifr[0].ifr_ifrn.ifrn_name, ifname, C.IFNAMSIZ)
        if C.ioctl(skfd, C.SIOCGMIIPHY, ifr) < 0 then
            local e = ffi.errno()
            C.close(skfd)
            error(string.format("SIOCGMIIPHY on %s failed: %s",  ifname,
                ffi.string(C.strerror(e))))
            return
        end
        local now = mdio_read(C.MII_BMSR)
        if bit.band(now, C.BMSR_LSTATUS) ~= 0 then
            return true
        end
    end

    local last_status = 'blah'

    ev.Timer.new(function(l, timer, e)
            local status = link_up(iface)
            if callback and last_status ~= status then
                last_status = status
                if callback(iface, status) then
                    C.close(skfd)
                    timer:stop(l)
                end
            end
        end, 1, 1):start(loop)
end

function net.wait_carrier(mode, noiface_mode, ifaces)
    local skfd = C.socket(C.AF_INET, C.SOCK_DGRAM, 0)
    if skfd < 0 then
        error("Unable to create socket")
    end

    local ifr = ffi.new('struct ifreq[1]')

    local function mdio_read(location)
        local mii = ffi.cast('struct mii_ioctl_data *',
            ifr[0].ifr_ifru.ifru_slave)
        mii.reg_num = location
        if C.ioctl(skfd, C.SIOCGMIIREG, ifr) < 0 then
            local e = ffi.errno()
            C.close(skfd)
            error(string.format("SIOCGMIIREG on %s failed: %s",
                ifr[0].ifr_ifrn.ifrn_name, ffi.string(C.strerror(e))))
        end
        return mii.val_out
    end

    -- Function used to determine if an interface exists.  Note, we do up to 10
    -- retries at a rate of 1/second on this because the ioctl can fail with
    -- EINVAL if this function is used early in the system boot sequence.
    local function iface_exists(ifname)
        for retry = 1, 10 do
            C.strncpy(ifr[0].ifr_ifrn.ifrn_name, ifname, C.IFNAMSIZ)
            if C.ioctl(skfd, C.SIOCGMIIPHY, ifr) < 0 then
                local e = ffi.errno()
                if e == C.ENODEV then
                    return false
                elseif e == C.EINVAL then
                    C.sleep(1)
                else
                    C.close(skfd)
                    error(string.format("SIOCGMIIPHY on %s failed: %s",  ifname,
                        ffi.string(C.strerror(e))))
                    return
                end
            end
            return true
        end
        error(string.format(
            "Unable to determine if interace %s exists", ifname))
    end

    local function link_up(ifname)
        C.strncpy(ifr[0].ifr_ifrn.ifrn_name, ifname, C.IFNAMSIZ)
        if C.ioctl(skfd, C.SIOCGMIIPHY, ifr) < 0 then
            local e = ffi.errno()
            C.close(skfd)
            error(string.format("SIOCGMIIPHY on %s failed: %s",  ifname,
                ffi.string(C.strerror(e))))
            return
        end
        local now = mdio_read(C.MII_BMSR)
        if bit.band(now, C.BMSR_LSTATUS) ~= 0 then
            return true
        end
    end

    local ifs = {}
    for i = 1, #ifaces do
        if not iface_exists(ifaces[i]) then
            print(string.format("No such network interface: %s", ifaces[i]))
            if noiface_mode == 'error' then
                return
            end
        else
            ifs[#ifs + 1] = ifaces[i]
        end
    end

    local function check_interfaces(ifs)
        local num_with_carrier = 0
        for i = 1, #ifs do
            if link_up(ifs[i]) then
                if mode == 'any' then
                    return true
                end
                num_with_carrier = num_with_carrier + 1
            end
        end

        -- We only get here when mode is not 'any'.  Return true only if all of
        -- the interfaces have carrier.
        if num_with_carrier == #ifs then
            C.close(skfd)
            return true
        end
    end

    -- If none of the specified interface have carrier wait, based on mode, for
    -- the carrier(s) on the interface(s) to come up.
    local lloop = net.ev.Loop.new()
    net.ev.Timer.new(
        function (loop, timer)
            if not check_interfaces(ifs) then
                return
            end
            C.close(skfd)
            timer:stop(loop)
        end, 1, 1):start(lloop)
    lloop:loop()
    C.close(skfd)
    return true
end

return net
