require 'strict'

local cjson = require 'cjson.safe'
local lpeg = require 'lpeg'
local sqlite = require 'sqlite_ffi'
local config = require 'system_config'

require 'middleclass'

local M = class('network_db')

local function dbg(fmt, ...)
    io.stderr:write(string.format(fmt, ...))
end

function M:initialize()
    self.db = assert(sqlite:new('/var/lib/armada/network.db'))
end

function M:close()
    self.db:close()
    self.db = nil
end

local function splitter(sep)
    sep = lpeg.P(sep)
    local elem = lpeg.C((1 - sep)^0)
    return lpeg.Ct(elem * (sep * elem)^0)
end

local splitters = {}
for _, sep in ipairs{'.', ' '} do
    splitters[sep] = splitter(sep)
end

local function split(s, sep)
    return lpeg.match(splitters[sep], s)
end

function M:query(q, ...)
    return self.db:exec(q, {...})
end

function M:begin_transaction()
    return self.db:exec('BEGIN TRANSACTION;')
end

function M:commit()
    return self.db:exec('COMMIT;')
end

function M:rollback()
    return self.db:exec('ROLLBACK;')
end

function M:query_owner_type(owner, type)
    local r = self:query([[
        SELECT rdata
        FROM unified
        WHERE owner = ? AND type = ?;
    ]], owner, type)
    return r
end

function M:available_channels(index_by)
    local o = {}
    local r = self:query[[
        SELECT substr(rdata, length('available_channels: ')) AS channels
        FROM unified
        WHERE type = 'TXT' AND rdata LIKE 'available*_channels: %' ESCAPE '*'
    ]]
    for i = 1, #r do
        local chs = cjson.decode(r[i].channels)
        if type(chs) == 'table' then
            for j = 1, #chs do
                if index_by and chs[j][index_by] then
                    o[chs[j][index_by]] = chs[j]
                else
                    o[#o + 1] = chs[j]
                end
            end
        end
    end
    return o
end

-- Function used to simple check to see if there are any tests running
function M:tests_running()
    self:begin_transaction()
    local r = self:query[[
        SELECT rdata FROM unified
        WHERE type = 'SRV' AND owner LIKE '%._armada-tests';
    ]]
    self:commit()
    if r and type(r) == 'table' and #r > 0 then return true end
end

-- Returns the port number to connect to running tests.  This is used to by
-- the power daemon to stop all running tests when the system is being brought
-- down.
function M:running_tests_port()
    local o = {}
    local r = self:query[[
        SELECT rdata AS info FROM unified
        WHERE type = 'SRV' AND owner LIKE '%*_armada-tests' ESCAPE '*'
            AND local_source <> '';
    ]]
    for i = 1, #r do
        local port = r[1].info:match('%d+ %d+ (%d+)')
        if port then
            o[#o + 1] = port
        end
    end
    return o
end

-- Generate a list of URL's that can be used to stop ALL running tests.
function M:get_stop_urls()
    local urls = {}
    local r = self:query[[
        SELECT rdata, source FROM unified
        WHERE type = 'SRV' AND owner LIKE '%*_armada-tests' ESCAPE '*'
    ]]
    for i = 1, #r do
        local port = r[i].rdata:match('%d+ %d+ (%d+)')
        local ip = r[i].source
        if port and ip then
            urls[#urls + 1] = string.format("http://[%s]:%s/stop", ip, port)
        end
    end
    return urls
end

-- Get the local host inside network hostname
function M:get_inside_network_hostname()
    local urls = {}
    local q = string.format(
        [[SELECT owner FROM unified
        WHERE type = 'A' AND owner LIKE '%s-%%' AND local_source <> '';]],
        config.device_type or 'cx23r')
    local r = self:query(q)
    if #r == 1 then
        return r[1].owner
    end
end

function M:resolve_name(name)
    local r = self:query_owner_type(name, 'CNAME')
    if r[1] then
        return r[1].rdata
    end

    local domain_parts = split(name, '.')
    local host_parts = { }
    table.insert(host_parts, table.remove(domain_parts, 1))
    while domain_parts[1] do
        local domain = table.concat(domain_parts, '.')
        local r = self:query_owner_type(domain, 'DNAME')
        if r[1] then
            local host = ''
            if #host_parts > 0 then
                host = table.concat(host_parts, '.') .. '.'
            end
            return self:resolve_hostname(host .. r[1].rdata)
        end
        table.insert(host_parts, table.remove(domain_parts, 1))
    end

    return name
end

function M:resolve_hostname(name)
    name = self:resolve_name(name)
    local r = self:query_owner_type(name, 'A')
    if r[1] then
        return r[1].rdata, r[1].owner
    end
    return nil, nil, 'Hostname not found'
end

function M:resolve_service(name)
    name = self:resolve_name(name)
    for i, rr in ipairs(self:query_owner_type(name, 'SRV')) do
        local port, host = rr.rdata:match('^%d+ %d+ (%d+) (.*)')
        local ip, canonical, err = self:resolve_hostname(host)
        if not ip then return ip, canonical, err end
        return host, port, ip, canonical
    end
    return nil, 'Service not found'
end

function M:resolve_alias(name)
    local rr = self:query_owner_type(name, 'ALIAS')
    if #rr > 0 then
        local subname = rr[1].rdata
        if subname == name then
            return name
        else
            return self:resolve_alias(rr[1].rdata)
        end
    end
    if name:find('%.') then
        local parts = split(name, '.')
        local h = table.remove(parts, 1) .. '.'
        local subname = table.concat(parts, '.')
        if subname == name then
            return h .. name
        else
            return h .. self:resolve_alias(subname)
        end
    end
    return name
end

local function get_local_info(self)
    if not self.local_info_cache then
        local r = self:query[[
            SELECT rdata
            FROM unified
            WHERE type = 'TXT'
                AND owner = '_armada-device-info'
                AND local_source <> '';
        ]]
        if #r == 1 and #r[1].rdata then
            self.local_info_cache = cjson.decode(r[1].rdata)
        end
    end
    return self.local_info_cache or {}
end

function M:get_module_hostname()
    return get_local_info(self).hostname
end

function M:get_module_ip()
    return get_local_info(self).internal_ip
end

function M:get_module_link_local_ip()
    return get_local_info(self).int_ll_ip
end

return M
