--! \file

require "strict"

local class = require "class"
local cjson = require "cjson"
local ev = require "ev"
local session_db = require "session_db"
local u = require "unix"
local uuid = require "uuid"

local temp_prefix = '/tmp/nginx/scgi'

local function hexify(s)
    return string.gsub(s, "([\r\n])", function(c)
        return string.format("%%%02X", string.byte(c))
    end)
end

--! An SCGI server.
--? class scgi_server server
local m = class(require "server")
--? alias m scgi_server

function m:init(settings)
    m.super().init(self, settings)
    self.session_db = session_db:new()
end

local function scgi_error(loop, w, message)
    u.write(w.socket, "HTTP/1.1 500 " .. message .. "\n\n")
    w.server.close(loop, w)
end

local function fixup_name(s)
    local r = s:lower():gsub('%-', '_')
    return r
end

local function fixup_value(s)
    local r = s:gsub('[ \t]*\r\n[ \t]+', ' ')
    return r
end

local function multipart_parsers(raw_boundary)
    local      C,      Cc,      Cf,      Cg,      Ct,      P,      R,      S
        = lpeg.C, lpeg.Cc, lpeg.Cf, lpeg.Cg, lpeg.Ct, lpeg.P, lpeg.R, lpeg.S
    local crlf = P'\x0D\x0A'
    local notcrlf = P(P(1) - crlf)
    local ows = P' '^0
    local sc = P';' * ows
    local header = Cg(P(R('AZ', 'az', '09', '--')^1 / fixup_name) * P':' * ows
        * P(notcrlf^0 * P(crlf * S' \t'^1 * notcrlf^0)^0 / fixup_value) * crlf)
    local boundary = P('--' .. raw_boundary)
    local item_parser = P(
        boundary * P(
            P'--' * crlf * Cc'' * lpeg.Cp()
          + P(crlf * Cf(Ct('') * header^0 * crlf, rawset)) * lpeg.Cp()
        )
    )
    local parsers = {
        start_item_parser = item_parser,
        item_parser = crlf * item_parser,
        item_end_parser = Ct(Cg(C(P(P(1) - crlf * boundary)^0), 'content'))
            * lpeg.Cp() * crlf * boundary,
        boundary_length = raw_boundary:len() + 4
    }
    return parsers
end

local states = {
    start = function(loop, w, frag)
        w.header_bytes = string.match(w.buf, "^(%d+):")
        if not w.header_bytes then
            if string.match(w.buf, "^%d*$") then
                return
            else
                scgi_error(loop, w, "Malformed SCGI request")
            end
        end
        w.header_len = tonumber(w.header_bytes)
        w.header_pos = w.header_bytes:len() + 2 -- skip ':'
        w.state = 'scgi_headers'
    end,
    scgi_headers = function(loop, w, frag)
        if w.buf:len() < w.header_pos + w.header_len then
            return
        end
        local pos = w.header_pos
        w.headers = {}
        while pos < w.header_len do
            local hdr, val = w.buf:match("^([^%z]*)%z([^%z]*)%z", pos)
            w.headers[hdr] = val
            pos = pos + hdr:len() + val:len() + 2
        end
        assert(pos == w.header_pos + w.header_len)
        assert(w.buf:sub(pos, pos) == ",")
        w.buf = w.buf:sub(pos + 1)
        -- KLUDGE error check below
        w.content_length = tonumber(w.headers['CONTENT_LENGTH'])
        w.content_values = {}
        w.content_type = w.server.decode_value_with_extras(w.headers.CONTENT_TYPE)
        if w.content_type.boundary then
            w.headers.multipart_boundary = w.content_type.boundary
            w.multiparsers = multipart_parsers(w.content_type.boundary)
            w.parser_name = 'start_item_parser'
            w.multipart = {}
            w.debug = {}
            w.state = 'incremental_multipart'
            w.reparse = true
            w.bytes_consumed = 0
            w.content = ''
        else
            w.headers.multipart_boundary = 'NOTHING'
            w.state = 'non_multipart'
        end
    end,
    non_multipart = function(loop, w, frag)
        w.headers.multipart = 'NOT IN multipart'
        -- KLUDGE need content reading handler
        if w.buf:len() >= w.content_length then
            w.content = w.buf:sub(1, w.content_length)
            w.buf = ''
            w.state = 'done'
        end
    end,
    incremental_multipart = function(loop, w, frag) -- multipart/form-data
        local part, pos = w.multiparsers[w.parser_name]:match(w.buf)
        local err
        if part == nil then
            local ok_len = w.buf:len() - w.multiparsers.boundary_length
            local mp = w.multipart[#w.multipart]
            if mp and mp.parsing and ok_len > 0 then
                local ok_buf = w.buf:sub(1, ok_len)
                if mp.tmp_fd then
                    local len = ok_buf:len()
                    local wrote = u.write(mp.tmp_fd, ok_buf)
                    if wrote ~= len then
                        err = "Error writing temp file"
                    end
                else
                    mp.content = mp.content .. ok_buf
                end
                w.buf = w.buf:sub(ok_len + 1)
                w.bytes_consumed = w.bytes_consumed + ok_len
            end
        elseif type(part) == 'table' then
            if part.content then
                local mp = w.multipart[#w.multipart]
                if mp.tmp_fd then
                    local wrote = u.write(mp.tmp_fd, part.content)
                    if wrote ~= #part.content then
                        err = "Error writing temp file"
                    end
                    if u.close(mp.tmp_fd) ~= 0 then
                        err = "Error closing temp file"
                    end
                    mp.tmp_fd = nil
                else
                    mp.content = mp.content .. part.content
                end
                mp.parsing = nil
                w.parser_name = 'item_parser'
            else
                part.content = ''
                part.parsing = true
                table.insert(w.multipart, part)
                local cdis = part.content_disposition
                if cdis then
                    local cd = w.server.decode_value_with_extras(cdis)
                    w.multipart['FOO_' .. #w.multipart] = cd
                    if cd then
                        part.content_disposition = nil -- KLUDGE ???
                        for k, v in pairs(cd) do
                            part[k] = v
                        end
                        if cd.filename then
                            w.files = w.files or {}
                            part.filename = cd.filename
                            part.tmp_filename, part.tmp_fd
                                = w.server:temp_file(w, '/tmp/nginx/scgi_')
                            if not part.tmp_fd then
                                err = "Error opening temp file"
                            end
                            -- KLUDGE ERROR CHECK (50X?) HERE!!!
                        end
                    end
                end
                w.parser_name = 'item_end_parser'
            end
            w.reparse = true
        elseif part == '' then
            w.state = 'done'
            w.reparse = true
        end
        if pos then
            w.bytes_consumed = w.bytes_consumed + pos - 1
            w.buf = w.buf:sub(pos)
        end
        if w.bytes_consumed >= w.content_length then
            w.state = 'done' -- KLUDGE ???
        end
    end,
    error = function(loop, w, frag)
        -- KLUDGE do something to propagate errors down here...
    end,
    done = function(loop, w, frag)
        w.api_internal = {}
        w.query = w.server.decode_query_string(w.headers.QUERY_STRING)
        w.query_string_values = w.query
        w.cookies = w.server.decode_cookies(w.headers.HTTP_COOKIE)
        w.accept_language
            = w.server.decode_accept_language(w.headers.HTTP_ACCEPT_LANGUAGE)
        if w.headers.HTTP_CONTENT_TYPE
                == 'application/x-www-form-urlencoded' then
            local cv = w.server.decode_query_string(w.content)
            for k, v in pairs(cv) do
                w.content_values[k] = v
                w.query[k] = v
            end
        end
        if w.multipart then
            for _, part in ipairs(w.multipart) do
                if part.name and part.content then
                    w.content_values[part.name] = part.content
                    w.query[part.name] = part.content
                end
            end
        end

        w.session_key = w.query.session_key
        w.authorization_key = w.cookies.authorization_key
            or w.query.authorization_key
        if w.session_key and w.server.session_db then
            w.session = w.server.session_db:get_session(w.session_key)
        end

        w.authorized = false
        if w.authorization_key then
            -- KLUDGE deal with expiration here ???
            local ses
            if w.session_key then
                ses = w.server.session_db:get_session(w.session_key)
                if ses and ses.authorization_key == w.authorization_key then
                    w.authorized = ses.authorization_key == w.authorization_key
                end
            end
            if not w.authorized then
                ses = w.server.session_db:get_session_from_authorization(
                    w.authorization_key)
                w.authorized = ses and true or false
                if w.authorized and w.server.auto_session then
                    w.session_key = uuid.generate()
                    local sec, nsec = u.clock_gettime()
                    w.session = {
                        session_key = w.session_key,
                        authorization_key = w.authorization_key,
                        timestamp = sec + nsec * 1e-9,
                        user_id = ses.user_id,
                        state = ses.state,
                    }
                    w.server.session_db:update_session(w.session)
                    ses = w.session
                end
            end
            if ses and w.authorized then
                w.session = ses
            end
        end

        w.server.handler(loop, w)
        w.state = nil
    end
}

function m.reader(loop, w, revents)
    local got = u.read(w.socket, 8192) -- KLUDGE increase for upload
    if not got or got == "" then
        w.server.close(loop, w)
        return
    end

    w.buf = w.buf .. got
    local buf_len = w.buf:len()

    local prev_state = ''
    w.reparse = false
    while w.state and (w.state ~= prev_state or buf_len ~= w.buf:len() or w.reparse) do

        prev_state = w.state
        buf_len = w.buf:len()
        w.reparse = false
        states[w.state](loop, w, got)
        got = ''
    end
end

function m.acceptor(loop, listen_w, revents)
    local w = m.super().acceptor(loop, listen_w, revents)
    if w then
        w.state = 'start'
        w.buf = ''
    end
end

function m.url_unescape(s)
    s = string.gsub(s, "+", " ")
    s = string.gsub(s, "%%(%x%x)", function (h)
        return string.char(tonumber(h, 16))
    end)
    return s
end

function m.decode_query_string(s)
    local q = {}
    if s then
        for name, value in string.gfind(s, "([^&=]+)=([^&=]+)") do
            q[m.url_unescape(name)] = m.url_unescape(value)
        end
    end
    return q
end

local value_with_extras_parser = (function()
    local C, Cf, Cg, Ct, P = lpeg.C, lpeg.Cf, lpeg.Cg, lpeg.Ct, lpeg.P
    local sc = P';' * P' '^0
    local quoted_value = P'"' * C(P(P(1) - P'"')^0) * P'"'
    local raw_value = C(P(P(1) - sc)^0)
    local value = quoted_value + raw_value
    local pair = Cg(C(P(P(1) - '=')^1) * '=' * value)
    return Cf(Ct("")
        * Cg(lpeg.Cc('value') * C(P(P(1) - ';')^1))
        * (sc * pair)^0 * sc^-1
        , rawset)
end)()

function m.decode_value_with_extras(s)
    local q = value_with_extras_parser:match(s)
    return q or {}
end

local cookie_parser = (function()
    local C, Cf, Cg, Ct, P = lpeg.C, lpeg.Cf, lpeg.Cg, lpeg.Ct, lpeg.P
    local sc = P';' * P' '^0
    local pair = Cg(C(P(P(1) - '=')^1) * '=' * C(P(P(1) - sc)^0))
    return Cf(Ct("") * pair * (sc * pair)^0 * sc^-1, rawset)
end)()

function m.decode_cookies(s)
    local q = {}
    if s then
        local raw = cookie_parser:match(s)
        for k, v in pairs(raw) do
            q[m.url_unescape(k)] = m.url_unescape(v)
        end
    end
    return q
end

local accept_language_parser = (function()
    -- "en-us,en;q=0.5"
    local      C,      Cc,      Cf,      Cg,      Ct,      P,      R
        = lpeg.C, lpeg.Cc, lpeg.Cf, lpeg.Cg, lpeg.Ct, lpeg.P, lpeg.R
    local alpha = R('az', 'AZ')
    local digit = R'09'
    local float = digit^1 * P'.' * digit^0 + P'.' * digit^1 + digit^1
    local ows = P' '^0
    local langlet = alpha^1 - alpha^9
    local language = Cg(P'*', 'base')
        + Cg(langlet, 'base') * P(P'-' * Cg(langlet, 'suffix'))^-1
    local quality = Cg((P';q=' * C(float) + Cc('1')) / tonumber, 'quality')
    local spec = Ct(language * quality)
    return Ct(ows * spec * P(ows * P',' * ows * spec * ows)^0)
end)()

function m.decode_accept_language(s)
    local q = {}
    if s then
        q = accept_language_parser:match(s)
        if q then
            for _, o in ipairs(q) do
                o.lang = o.base .. (o.suffix and ('-' .. o.suffix) or '')
            end
            table.sort(q, function(e1, e2) return e1.quality > e2.quality end)
        end
    end
    return q
end

local function write(socket, out)
    local len = out:len()
    while true do
        local amt, estr, eno = u.write(socket, out)
        if not amt then
            if eno ~= u.EAGAIN then return end
        elseif amt == out:len() then break
        elseif amt == 0 then break
        else out = string.sub(out, amt + 1)
        end
    end
    return len
end

function m.begin_comet(self, w, comet_function)
    local content_type = "text/html; charset=utf-8"
    local user_agent = string.lower(w.headers["HTTP_USER_AGENT"])
    if string.match(user_agent, "android") then
        w.is_android = 1
    elseif string.match(user_agent, "webkit") then
        -- KLUDGE! See http://stackoverflow.com/questions/169258 \
        --   /is-http-streaming-comet-possible-in-safari
        -- Also, Android matches webkit but wants text/html
        content_type = "multipart/x-mixed-replace"
    end
    if not comet_function then
        comet_function = "window.parent.comet_update"
    end

    local out =
        "HTTP/1.1 200 OK\n"
        .. "Content-Type: " .. content_type .. "\n"
        .. "Content-Script-Type: text/javascript\n"
        .. "Cache-Control: private\n"
        .. "Pragma: no-cache\n"
        .. "X-Accel-Buffering: no\n"
        .. "\n"
        -- Workaround for IE: see http://forums.devx.com/showthread.php?t=163336
        ..  string.rep(" ", 256)
        .. "<script>\n"
        .. "function C(data) {\n"
        .. "    var scripts = document.getElementsByTagName('script');\n"
        .. "    var script = scripts[scripts.length - 1];\n"
        .. "    script.parentNode.removeChild(script);\n"
        .. "    for (var prop in script) {\n"
        .. "        delete script[prop];\n"
        .. "    }\n"
        .. "    " .. comet_function .. "(data);\n"
        .. "}\n"
        .. "</script>\n"

    write(w.socket, out)
end

function m.begin_eventsource(self, w, comet_function)
    local content_type = "text/event-stream"

    local out =
        "HTTP/1.1 200 OK\n"
        .. "Content-Type: " .. content_type .. "\n"
        .. "Content-Script-Type: text/javascript\n"
        .. "Cache-Control: private\n"
        .. "Pragma: no-cache\n"
        .. "X-Accel-Buffering: no\n"
        .. "\n"

    write(w.socket, out)
    w.eventsource = true
end

function m.write_comet(self, w, json)
    if type(json) == "table" then
        json = cjson.encode(json)
    end
    if w.eventsource then
        local out = string.format("data: %s\n\n", json)
        write(w.socket, out)
    else
        local out = string.format("<script>\nC(%s)\n</script>\n", json)
        local len = write(w.socket, out)
        if len and w.is_android and len < 4096 then
            -- KLUDGE Android browser buffers about 4k
            -- of input before updating
            write(w.socket, string.rep(" ", 4096 - len) .. "\n")
        end
    end
end

return m
