local cjson         = require 'cjson'
local ffi           = require 'ffi'
local lfs           = require 'lfs'
local sqlite        = require 'sqlite_ffi'
local system_config = require 'system_config'
local uuid          = require 'uuid'
local log           = require 'log'

local C = ffi.C

local m = {}

--m.setup_db = sqlite:new('/var/lib/armada/setup.db')
m.storage_path = system_config.storage.path or '/storage'

local function sql_exec(db, table_name, sql, args)
    local query = string.format(sql, table_name):gsub('^%s*(.-)%s$', '%1')
    local ok, result = pcall(function() return db:exec(query, args) end)
    if ok then return result, ok
    else return { ok = false, error = result }, ok end
end

-- Get the list of SIE files.  This is basically a mapping between the SIE file
-- name as stored on the storage pack and the "friendly" name assigned by the
-- user.

function m.get_sie_file_list()
    m.setup_db = m.setup_db or sqlite:new('/var/lib/armada/setup.db')

    local result = {}
    local r, ok = sql_exec(m.setup_db, 'tsd_users',
        [[SELECT content, id FROM "%s"
        WHERE db_name == 'sie_name' AND deleted == '';]])
    if ok then
        for _, v in ipairs(r) do
            if v.content and v.id then
                local c = cjson.decode(v.content)
                if c then
                    local stat =
                        lfs.attributes(m.storage_path .. '/' .. v.id .. '.sie')
                    if stat then
                        result[#result + 1] = {
                            name = c.friendly_name,
                            size = stat.size,
                            access_time =
                                os.date('!%FT%H:%M:%S', stat.access),
                            change_time =
                                os.date('!%FT%H:%M:%S', stat.change),
                            modification_time =
                                os.date('!%FT%H:%M:%S', stat.modification),
                            author = c.author,
                            summary = c.summary
                        }
                    end
                end
            end
        end
    end
    return result
end

-- This function returns the path to the actual SIE file associated with the
-- "friendly" name given.

function m.get_sie_path_on_disk(name)
    m.setup_db = m.setup_db or sqlite:new('/var/lib/armada/setup.db')
    local r, ok = sql_exec(m.setup_db, 'tsd_users',
        [[SELECT content, id FROM "%s"
        WHERE db_name == 'sie_name' AND deleted = '';]])
    if ok then
        for _, v in ipairs(r) do
            if v.content and v.id then
                local c = cjson.decode(v.content)
                if c then
                    if c.friendly_name and c.friendly_name == name then
                        return m.storage_path .. '/' .. v.id .. '.sie'
                    end
                end
            end
        end
    end
end

-- This function removes the association between a "friendly" name and an SIE
-- file on the disk.

function m.delete_sie_friendly_name(name)
    m.setup_db = m.setup_db or sqlite:new('/var/lib/armada/setup.db')
    local path_on_disk = m.get_sie_path_on_disk(name)
    if not path_on_disk then return end
    local txt_file = path_on_disk:gsub('%.sie', '%.txt')
    local id = path_on_disk:match(m.storage_path .. '/(.+)%.sie')
    local result = {}
    local request = string.format(
        [[DELETE FROM "%%s" WHERE db_name == 'sie_name' AND id == '%s';]], id)
    local r, ok = sql_exec(m.setup_db, 'tsd_users', request)
    C.unlink(txt_file)
    return path_on_disk
end

-- This function removes the association between a "friendly" name and an SIE
-- file on the disk AND removes the SIE from the disk.

function m.delete_sie_file(name)
    local path_on_disk = m.delete_sie_friendly_name(name)
    if path_on_disk then
        C.unlink(path_on_disk)
        return true
    end
end

function m.delete_sie_files(list)
print("delete_sie_files")
    -- Get the ID to friendly name mappings
    m.setup_db = m.setup_db or sqlite:new('/var/lib/armada/setup.db')
    local request = string.format(
        [[SELECT content, id FROM "%%s" WHERE db_name == 'sie_name' AND
        deleted = '' AND content LIKE '%%%%friendly_name":"%%%%';]])
    local ids, ok = sql_exec(m.setup_db, 'tsd_users', request)
    if not ok then return {}, list end

    for i, v in ipairs(ids) do
        if not v.content or not v.id then
            ids[i] = nil
        else
            ids[i].decoded = cjson.decode(v.content)
            if not ids[i].decoded then
                ids[i] = nil
            end
        end
    end

    local function remove_it(name)
        name = name:gsub('%.sie', '')
        for i, v in ipairs(ids) do
            if v.decoded.friendly_name == name then
                local request = string.format(
                    [[DELETE FROM "%%s" WHERE db_name == 'sie_name'
                    AND id == '%s';]], v.id)
                sql_exec(m.setup_db, 'tsd_users', request)
                C.unlink(m.storage_path .. '/' .. v.id .. '.txt')
                C.unlink(m.storage_path .. '/' .. v.id .. '.sie')
                return i
            end
        end
    end

    -- Walk the list of names removing the SIE files
    local removed, nonremoved = {}, {}
    for _, name in ipairs(list) do
        local i = remove_it(name)
        if i then
            table.remove(ids, i)
            removed[#removed + 1] = name
        else
            nonremoved[#nonremoved + 1] = name
        end
    end
    return removed, nonremoved
end

-- This function associates a "friendly" name with an SIE file on the disk.
-- If the friendly name already exists, a new one will be generated.  The name
-- used will be returned.

function m.assign_friendly_sie_name(requested_name, path_on_disk, author,
    summary)
    m.setup_db = m.setup_db or sqlite:new('/var/lib/armada/setup.db')
    local req = string.format(
        [[SELECT content FROM "%%s" WHERE db_name = 'sie_name' AND
        deleted = '' AND content LIKE '%%%%friendly_name":"%s%%%%'
        ORDER BY rowid;]], requested_name)
    local r, ok = sql_exec(m.setup_db, 'tsd_users', req)
    if ok then
        local map
        for _, v in ipairs(r) do
            local c = cjson.decode(v.content)
            if c.friendly_name == requested_name then
                map = map or {}
                map[0] = requested_name
            else
                local num = c.friendly_name:match('%-(%d+)$')
                if num then
                    num = tonumber(num)
                    if num then
                        map = map or {}
                        map[num] = true
                    end
                end
            end
        end
        local name = requested_name
        if map then
            if #map + 1 < 1000 then
                name = string.format("%s-%03d", requested_name, #map + 1)
            else
                name = string.format("%s-%d", requested_name, #map + 1)
            end
        end

        -- Add the name mapping to the database
        local info = {
            friendly_name = name,
            author = author,
            summary = summary
        }
        local id = path_on_disk:match(m.storage_path .. '/(.+)%.sie')
        sql_exec(m.setup_db, 'tsd_users',
            [[INSERT INTO "%s" (db_name, id, content, added)
            VALUES ('sie_name', ?, ?, ?);
        ]], {id, cjson.encode(info), os.date('!%FT%H:%M:%S', os.time())})

        local f = io.open(path_on_disk:gsub('%.sie$', '%.txt'), 'w')
        if f then
            f:write(name)
            f:close()
        end
        log:log(log.NOTICE, string.format("Selected 'friendly' SIE name: '%s'",
            name))
        return name
    end
end

-- This function changes the "friendly" name of an SIE file and returns the new
-- friendly name.

function m.rename_sie(requested_name, current_name)
    local params = m.get_sie_parameters_by_name(current_name)
    local pod = m.delete_sie_friendly_name(current_name)
    if pod then
        local new_pod = pod:match('^(/.+/)') .. uuid.generate() .. '.sie'
        if new_pod then
            local assigned_name =
                m.assign_friendly_sie_name(requested_name, new_pod,
                    params.author, params.summary)
            if assigned_name then
                os.rename(pod, new_pod)
                return assigned_name
            end
        end
        -- Any failure restores the old mapping
        m.assign_friendly_sie_name(current_name, pod, params.author,
            params.summary)
    end
end

-- This function is used to clean out friendly name to actual file associations
-- from the database for friendly names whos actual file no longer exists.

function m.clean_sie()
    m.setup_db = m.setup_db or sqlite:new('/var/lib/armada/setup.db')
    local result = {}
    local r, ok = sql_exec(m.setup_db, 'tsd_users',
        [[SELECT content, id FROM "%s"
        WHERE db_name == 'sie_name' AND deleted == '';]])
    if ok then
        for _, v in ipairs(r) do
            if v.content and v.id then
                if not lfs.attributes(
                    m.storage_path .. '/' .. v.id .. '.sie') then
                    local req = string.format(
                        [[DELETE FROM "%%s" WHERE db_name == 'sie_name'
                        AND id == '%s';]], v.id)
                    sql_exec(m.setup_db, 'tsd_users', req)
                end
            end
        end
    end
    return result
end

-- This function is used to get the "friendly" name associated with a
-- particular SIE file.

function m.get_sie_parameters_by_hash(hash_name)
    m.setup_db = m.setup_db or sqlite:new('/var/lib/armada/setup.db')
    hash_name = string.gsub(hash_name, '%.sie', '')
    local req = string.format(
        [[SELECT content FROM "%%s"
        WHERE db_name == 'sie_name' AND deleted == '' AND id == '%s';]],
        hash_name)
    local r, ok = sql_exec(m.setup_db, 'tsd_users', req)
    if ok then
        for _, v in ipairs(r) do
            if v.content then
                return cjson.decode(v.content)
            end
        end
    end
end

-- This function is used to get the "friendly" name associated with a
-- particular SIE file.

function m.get_sie_parameters_by_name(name)
    m.setup_db = m.setup_db or sqlite:new('/var/lib/armada/setup.db')
    local req = [[SELECT content FROM "%s"
        WHERE db_name == 'sie_name' AND deleted == '';]];
    local r, ok = sql_exec(m.setup_db, 'tsd_users', req)
    if ok then
        for _, v in ipairs(r) do
            if v.content then
                local x = cjson.decode(v.content)
                if x.friendly_name == name then
                    return x
                end
            end
        end
    end
end

-- Function used to check the SIE hash name to friendly name mappings

function m.check_sie_files()
    -- Process all of the SIE files in the storage path
    for fh in lfs.dir(m.storage_path) do
        -- Is this an SIE file ?
        if fh:match('%.sie$') then
            -- Yes, get the existing hash to friendly mapping from the database
            local params = m.get_sie_parameters_by_hash(fh)
            local fn = params and params.friendly_name or nil

            -- Generate the path to the backup mapping file
            local bu = m.storage_path .. '/' ..  fh:gsub('%.sie', '%.txt')

            -- Is there an entry in the database for this SIE file?
            if not fn then
                -- Nope, is there a backup name mapping file for it?
                local fi = io.open(bu, 'r')
                if fi then
                    -- Yes, get the mapping
                    fn = fi:read('*a'):gsub('.*/', '')
                    fi:close()
                else
                    -- No backup mapping of friendly name to hash name, attempt
                    -- to use the test name in the SIE file
                    -- KLUDGE:  This should probably be done by actuall reading
                    -- and parsing the XML for the information.
                    fn = fh:match('(.+)%.sie$')
                    local f = io.popen('grep -m 1 \'"test_name": "\' ' ..
                        m.storage_path .. '/' .. fh)
                    if f then
                        for l in f:lines() do
                            fn = l:match('"test_name": "(.+)",')
                            if fn then
                                break
                            end
                        end
                        f:close()
                    end
                end

                -- Add the mapping to the database
                fn = m.assign_friendly_sie_name(fn, m.storage_path .. '/' .. fh)
            end
            -- Make sure backup mapping is set properly.
            local f = io.open(bu, 'w')
            if f then
                f:write(fn)
                f:close()
            end
        end
    end

    -- Clean out all friendly name mapping files for which the associated SIE
    -- no longer exists
    for fh in lfs.dir(m.storage_path) do
        -- Is this an friendly name mapping file?
        -- FIXME:  The pattern is not selective enough to prevent removal of
        -- all TXT files in the directory.
        if fh:match('%.txt$') then
            local sie_path = m.storage_path .. '/' ..  fh:gsub('%.txt', '%.sie')
            local stat = lfs.attributes(sie_path)
            if not stat then
                -- No associated SIE file, remove the TXT file
                C.unlink(m.storage_path .. '/' .. fh)
            end
        end
    end
end

return m
