require 'strict'

-- down_sampler.lua - Implements the Down Sampler computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local sample_clock = require 'sample_clock'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof
local ffi_copy = ffi.copy

local math_ceil = math.ceil
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Frame_tail = frame.tail
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_has_frames = pipe.has_frames
local Pipe_get_frame = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}

function mt:dump()
    dbg('factor=%lld', self.factor)
end

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

-- Processes incoming frames, produces outgoing frames.
-- NOTE that output frames will be smaller than input frames
-- by self.factor !!!
function mt:run()
    if not Pipe_has_frames(self.in_pipe) then
        return
    end

    local out_frame_size =
        frame.bufsize(self.sample_rate, self.pipe_rate)

    local f = Pipe_get_frame(self.in_pipe)
    while f ~= nil do
        for i = 0, f.count - 1 do
            if self.tot_input_count % self.factor == 0 then
                if self.current_frame == nil then
                    self.current_frame =
                        Frame_create(out_frame_size, f.element_size)
                    self.current_frame.hdr.ts = self.tot_count
                    self.current_frame.ts = math.floor(self.tot_count)
                end
                local of = self.current_frame
                ffi_copy(Frame_tail(of),
                    ffi_cast('char *', f.buf) + (f.element_size * i),
                    f.element_size)
                of.count = of.count + 1
                if of.count == of.maxsize then
                    self.tot_count = self.tot_count + of.count
                    if not utils.put_frame(of, self.pipe_list) then
                        Frame_delete(of)
                    end
                    self.current_frame = nil
                end
            end
            self.tot_input_count = self.tot_input_count + 1
        end
        Frame_delete(f)
        f = Pipe_get_frame(self.in_pipe)
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local down_sampler = setmetatable({}, {__index = mt})
    down_sampler.sie_collect_idx = first_sie_index
    utils.set_common_meta(down_sampler, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Down Sampler computed channel only" ..
        " accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    down_sampler.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(down_sampler.pipe_rate and down_sampler.pipe_rate > 0,
        "Invalid frame rate: " .. down_sampler.pipe_rate)
    --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    down_sampler.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(down_sampler.sample_rate and down_sampler.sample_rate > 0,
        "Invalid sample rate: " .. down_sampler.sample_rate)
    --   down sampler 'factor'
    down_sampler.factor = tonumber(stanza.factor)
    assert(down_sampler.factor and down_sampler.factor >= 1,
        "INVALID down sample factor: " .. tostring(down_sampler.factor))
    
    -- create input/output pipe structures
    down_sampler.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(down_sampler.in_pipe and down_sampler.in_pipe ~= nil,
        "failed to create input pipe")
    down_sampler.pipe_list = Pipe_new_pipe_list()
    assert(down_sampler.pipe_list and down_sampler.pipe_list ~= nil,
        "failed to create (output) pipe list")

    -- specifies the first input sample to be put in the output
    -- data - will be updated after each input frame
    down_sampler.first_sample_index = 0
    -- the total of all input samples, used to update 'first_sample_index'
    down_sampler.tot_input_count = 0

    -- used for frames' time stamps (just a sample count for
    -- this module)
    down_sampler.tot_count = 0

    return down_sampler, 1
end

return m
